package ca.ualberta.cs.poker.free.tournament;

import java.net.InetAddress;
import java.util.Vector;
import java.security.SecureRandom;

import ca.ualberta.cs.poker.free.dynamics.LimitType;

/**
 * Runs a round-robin tournament of either heads-up limit or
 * heads-up no-limit. 
 * At present, only limit
 * @author Martin Zinkevich
 *
 */
public class RoundRobinTournament implements Node{
  /**
   * Bots in the tournament.
   */
  Vector<BotInterface> bots;
  
  /**
   * Whether the game is limit or no-limit.
   * No-limit also implies stackBoundGame
   */
  boolean limit;
  
  /**
   * Series between bots in the tournament.
   */
  Vector<HeadsUpSeries> series;


  /**
   * Root of all series names.
   */
  String rootSeriesName;
  
  /**
   * Root card file name.
   */
  String rootCardFileName;

  /** 
   * The number of duplicate match pairs per series.
   */
  int numDuplicatePairs;

  /**
   * The server IP.
   */
  InetAddress server;
  
  /**
   * Construct a tournament from a bot list,
   * root names, number of duplicate match pairs per series,
   * and server IP.
   */
  public RoundRobinTournament(Vector<BotInterface> bots,
  String rootSeriesName, String rootCardFileName, int
  numDuplicatePairs,InetAddress server, boolean limitGame){
    this.bots = bots;
    this.rootSeriesName = rootSeriesName;
    this.rootCardFileName = rootCardFileName;
    this.numDuplicatePairs = numDuplicatePairs;
    this.server = server;
    initHeadsUpLimitSeries();
  }

  /**
   * Create the series objects.
   */
  public void initHeadsUpLimitSeries(){
    series = new Vector<HeadsUpSeries>();

    for(int i=0;i<bots.size();i++){
      for(int j=i+1;j<bots.size();j++){
        BotInterface playerA = bots.get(i);
	BotInterface playerB = bots.get(j);
	String rootMatchName = rootSeriesName + "."+
	playerA.getName()+"."+playerB.getName()+".match";
	series.add(new HeadsUpSeries(playerA,playerB,rootMatchName,
	rootCardFileName,numDuplicatePairs,server,LimitType.LIMIT));
      }
    }
  }

  public void showStatistics(){
    for(HeadsUpSeries s:series){
      s.showStatistics();
    }
  }
  
  /**
   * Get the utilities from the round robin tournament.
   */
  public Vector<Integer> getUtilities(){
    int[] result = new int[bots.size()];
    for(HeadsUpSeries s:series){
      Vector<BotInterface> seriesBots = s.getBots();
      Vector<Integer> seriesUtils = s.getUtilities();
      for(int i=0;i<seriesBots.size();i++){
        int botIndex = bots.indexOf(seriesBots.get(i));
	result[botIndex]+=seriesUtils.get(i);
      }
    }
    Vector<Integer> finalResult = new Vector<Integer>();
    for(int i=0;i<result.length;i++){
      finalResult.add(result[i]);
    }
    return finalResult;
  }
      
  public boolean isComplete(){
    //System.out.println("roundRobinTournament.isComplete()");
    for(HeadsUpSeries s:series){
      if (!s.isComplete()){
        //System.out.println("roundRobinTournament.isComplete() finished");
        return false;
      }
    }
    //System.out.println("roundRobinTournament.isComplete() finished");
    return true;
  }

  public void load(Forge f){
    for(HeadsUpSeries s:series){
      s.load(f);
    }
  }

  /**
   * Get the winner of the tournament.
   */
  public BotInterface getWinner(){
    Vector<Integer> utils = getUtilities();
    int maxSoFar = utils.get(0);
    BotInterface botSoFar = bots.get(0);

    for(int i=1;i<utils.size();i++){
      if (maxSoFar<utils.get(i)){
        botSoFar = bots.get(i);
	maxSoFar = utils.get(i);
      }
    }
    return botSoFar;
  }

  public void generateCardFiles(SecureRandom random){
    for(HeadsUpSeries s:series){
      s.generateCardFiles(random);
    }

  }

  /**
   * Show the matches and the bots in the tournament.
   */
  public String toString(){
    String result = "Matches:\n";
    for(HeadsUpSeries s:series){
      result+=s.toString();
    }
    result+="Bots:\n";
    for(BotInterface b:bots){
      result+=(b.toString()+"\n");
    }
    return result;
  }

}
